/*
 Copyright 2016 Pagemesh Inc.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */

module.exports = function(log, _, db$_, schema$_, ResourceContainer, injector) {

    const expressRouter = require('express-promise-router');

    const PHASES = {
        BEFORE_CREATE: 'before-create',
        AFTER_CREATE: 'after-create',
        BEFORE_UPDATE: 'before-update',
        AFTER_UPDATE: 'after-update',
        BEFORE_REMOVE: 'before-remove',
        AFTER_REMOVE: 'after-remove',
        BEFORE_LIST: 'before-list',
        AFTER_LIST: 'after-list',
        BEFORE_READ: 'before-read',
        AFTER_READ: 'after-read'
    };

    class Resource {

        constructor(opts = {}) {

            // Private members
            this._hooks = {};
            this._containers = {};

            // Public members
            this.key = opts.key;
            this.auth = opts.auth;
            this.cache = opts.cache;
            this.description = opts.description;

            this.collection = opts.collection || this.key;

            if (opts.db) {
                this.db = opts.db;
                if (_.isString(opts.db)) {
                    this.db = db$_[opts.db];
                }
            } else {
                this.db = db$_.main;
            }

            // Schema can be provided as a string or a resolved schema object
            if (opts.schema) {
                this.schema = opts.schema;
                if (_.isString(this.schema)) {
                    this.schema = schema$_[opts.schema];
                }
            }
        }

        /**
         * Connect a custom hook to a specific resource phase. 
         *  
         * @param  {PHASES} phase
         * @param  {Function} handler
         */
        addHook(phase, handler, opts = {
            priority: 5
        }) {

            // Lazy initialization
            if (!this._hooks[phase]) {
                this._hooks[phase] = [];
            }

            // Add this hook
            this._hooks[phase].push({
                handler: handler,
                priority: opts.priority
            });
        }

        /**
         * @param  {string} name
         * @param  {Resource} resource
         */
        addContainer(name, resource) {
            this._containers[name] = new ResourceContainer(this, resource);
            return this._containers[name];
        }

        /**
         * Retrieve a registered container for this resource 
         * 
         * @param {string} name The name of the container as registered
         * @returns {ResourceContainer} The resource container or undefined
         */
        getContainer(name) {
            return this._containers[name];
        }

        get router() {
            let router = expressRouter();

            let middleware = [];

            // Connect Security
            if (this.auth) {
                let passport = injector.resolve('passport');
                if (passport) {
                    if (_.isFunction(this.auth)) {
                        this.auth.call(this, middleware);
                    } else if (this.auth.type) {
                        middleware.push(passport.authenticate(this.auth.type, this.auth.options));
                    } else {
                        middleware.push(passport.authenticate(this.auth));
                    }
                }
            }

            // TODO: Connect Caching

            router.get("/", middleware, injector.execute(require('./handlers/list.handler'), this));
            router.get("/:id", middleware, injector.execute(require('./handlers/read.handler'), this));
            router.put("/:id", middleware, injector.execute(require('./handlers/update.handler'), this));
            router.patch("/:id", middleware, injector.execute(require('./handlers/patch.handler'), this));
            router.post("/", middleware, injector.execute(require('./handlers/create.handler'), this));
            router.delete("/:id", middleware, injector.execute(require('./handlers/remove.handler'), this));

            // Register all containers
            _.forEach(this._containers, (val, key) => {
                router.use(`/${key}`, val.router);
            });

            return router;
        }

        // All application code should use these constants
        static get PHASES() {
            return PHASES;
        }
    }

    return Resource;
};